@@ -1,9 +1,5 @@
-# https://gradio.app/blocks-and-event-listeners/
-# demo: http://10.147.17.2:7860/
-
 import json
 import os
-import pdb
 from operator import itemgetter
 from pathlib import Path
 
@@ -12,11 +8,15 @@ import gradio as gr
 import numpy as np
 import openai
 import pandas as pd
-from sentence_transformers import SentenceTransformer, util
-from tqdm import tqdm
+import torch
+import torch.nn.functional as F
+from transformers import AutoModel, AutoTokenizer
 
+from . import util_funcs
 from .zca import ZCA
 
+device = "cuda" if torch.cuda.is_available() else "cpu"
+
 
 class CVFDemo:
     # example values used in the paper
@@ -66,9 +66,10 @@ class Meaurement:
         self.scale_definitions = {}
 
     def read_csv_cols(self, file_obj):
-        self.input_df = pd.read_csv(file_obj[0].name)
-        self.col_names = self.input_df.columns.tolist()
-        print(f"Columns: {self.col_names}")
+        if file_obj is not None:
+            self.input_df = pd.read_csv(file_obj[0].name, encoding="ISO-8859-1")
+            self.col_names = self.input_df.columns.tolist()
+            print(f"Columns: {self.col_names}")
         return gr.Dropdown.update(
             choices=self.col_names, interactive=True
         ), gr.Dropdown.update(choices=self.col_names, interactive=True)
@@ -96,7 +97,12 @@ class Meaurement:
     def set_sbert(self, model_name):
         print(f"Setting SBERT model to {model_name}")
         self.model_name = model_name
-        self.model = SentenceTransformer(model_name)
+        self.tokenizer = AutoTokenizer.from_pretrained(
+            f"sentence-transformers/{model_name}"
+        )
+        self.model = AutoModel.from_pretrained(f"sentence-transformers/{model_name}")
+        self.model.eval()
+        self.model.to(device)
 
     def set_openai_api_key(self, api_key):
         print("Setting OpenAI API key")
@@ -159,17 +165,40 @@ class Meaurement:
         if embedding_option == "Embed Documents":
             return gr.File.update(visible=False), gr.Button.update(visible=True)
 
-    def embed_texts(self, sentences, progress=gr.Progress()):
+    def embed_texts(self, sentences, progress, batch_size=8):
+        """Use huggingface transformers to embed the text_col"""
+        sentences = [str(x) for x in sentences]
         if self.use_openAI is False:
             # use sentence_transformers to embed the text_col
             print("Embedding Using Sentence Transformers")
-            sentence_embeddings = self.model.encode(
-                sentences, show_progress_bar=True, normalize_embeddings=True
-            )
+            # use batching
+            with torch.no_grad():
+                self.model.eval()
+                sentence_embeddings = []
+                for i in progress.tqdm(
+                    range(0, len(sentences), batch_size), unit="batch (batch size=8)"
+                ):
+                    batch = sentences[i : i + batch_size]
+                    encoded_input = self.tokenizer(
+                        batch, padding=True, truncation=True, return_tensors="pt"
+                    ).to(device)
+                    model_output = self.model(**encoded_input)
+                    batch_sentence_embeddings = util_funcs.mean_pooling(
+                        model_output, encoded_input["attention_mask"]
+                    )
+                    batch_sentence_embeddings = F.normalize(
+                        batch_sentence_embeddings, p=2, dim=1
+                    )
+                    sentence_embeddings.append(batch_sentence_embeddings)
+                sentence_embeddings = torch.cat(sentence_embeddings, dim=0)
+                # to cpu and numpy
+                sentence_embeddings = sentence_embeddings.cpu().detach().numpy()
+                assert sentence_embeddings.shape[0] == len(sentences)
+
         else:
             print("Embedding Using OpenAI")
             sentence_embeddings = []
-            for sent in tqdm(sentences):
+            for sent in progress.tqdm(sentences, unit="docs"):
                 response = openai.Embedding.create(
                     input=sent, model="text-embedding-ada-002"
                 )
@@ -183,16 +212,28 @@ class Meaurement:
 
         return sentence_embeddings
 
-    def embed_df(self):
-        self.embeddings = self.embed_texts(self.input_df[self.doc_col_name])
-        np.save(Path(self.path_mgt.out_dir, "embeddings.npy"), self.embeddings)
-        return (
-            gr.Textbox.update(
-                value=f"Embedding completed! \nThe shape of the embeddings is {self.embeddings.shape}. You can download and save the embeddings below. Proceed to the next tab to define the dimensions."
-            ),
-            gr.File.update(visible=True),
-            Path(self.path_mgt.out_dir, "embeddings.npy"),
-        )
+    def embed_df(self, progress=gr.Progress()):
+        try:
+            self.embeddings = self.embed_texts(
+                self.input_df[self.doc_col_name], progress=progress
+            )
+            np.save(Path(self.path_mgt.out_dir, "embeddings.npy"), self.embeddings)
+            if (self.doc_col_name is not None) and (
+                Path(self.path_mgt.out_dir, "embeddings.npy").exists()
+            ):
+                return (
+                    gr.Textbox.update(
+                        value=f"Embedding completed! \nThe shape of the embeddings is {self.embeddings.shape}. You can download and save the embeddings below. Proceed to the next tab to define the dimensions."
+                    ),
+                    gr.File.update(visible=True),
+                    Path(self.path_mgt.out_dir, "embeddings.npy"),
+                )
+        except Exception as e:
+            return (
+                gr.Textbox.update(value=f"Embedding failed! \nError message: {e}"),
+                gr.File.update(visible=False),
+                None,
+            )
 
     def init_query_boxes(self, n_dims):
         # ! deprecated
@@ -242,140 +283,176 @@ class Meaurement:
                 update_rows.append(gr.Row.update(visible=False))
         return update_rows
 
-    def semantic_search(self, query, n_results):
+    def semantic_search(self, query, n_results, progress=gr.Progress()):
         # search for the most similar documents
         result_str = ""
         queries = query.split("\n")
         # remove empty queries or queries with only spaces
         queries = [q for q in queries if q.strip() != ""]
-        query_embedding = self.embed_texts(queries)
-        mean_vect_query = query_embedding.mean(axis=0)
-        hits = util.semantic_search(
-            mean_vect_query,
-            self.embeddings,
-            score_function=util.dot_score,
-            top_k=n_results,
-        )
-        hit_ids = [h["corpus_id"] for h in hits[0]]
-        hit_scores = [h["score"] for h in hits[0]]
-        examples = itemgetter(*hit_ids)(self.input_df[self.doc_col_name])
-        examples = list(examples)
-        example_doc_ids = itemgetter(*hit_ids)(self.input_df[self.doc_id_col_name])
-        example_doc_ids = list(example_doc_ids)
-        assert len(hit_ids) == len(hit_scores) == len(examples)
-        for i in range(len(hit_ids)):
-            result_str += f"Document ID: {example_doc_ids[i]} \n"
-            result_str += f"Score: {round(hit_scores[i], 3)} \n"
-            result_str += "------------------------ \n"
-            result_str += f"{examples[i]}\n"
-            result_str += "------------------------ \n"
+        try:
+            query_embedding = self.embed_texts(queries, progress=progress)
+            mean_vect_query = query_embedding.mean(axis=0)
+            hits = util_funcs.semantic_search(
+                mean_vect_query,
+                self.embeddings,
+                score_function=util_funcs.dot_score,
+                top_k=n_results,
+            )
+            hit_ids = [h["corpus_id"] for h in hits[0]]
+            hit_scores = [h["score"] for h in hits[0]]
+            examples = itemgetter(*hit_ids)(self.input_df[self.doc_col_name])
+            examples = list(examples)
+            example_doc_ids = itemgetter(*hit_ids)(self.input_df[self.doc_id_col_name])
+            example_doc_ids = list(example_doc_ids)
+            assert len(hit_ids) == len(hit_scores) == len(examples)
+            for i in range(len(hit_ids)):
+                result_str += f"Document ID: {example_doc_ids[i]} \n"
+                result_str += f"Score: {round(hit_scores[i], 3)} \n"
+                result_str += "------------------------ \n"
+                result_str += f"{examples[i]}\n"
+                result_str += "------------------------ \n"
+        except Exception as e:
+            result_str = (
+                f"Error. Make sure the query is not empty. \nError message: {e}"
+            )
         return gr.Textbox.update(value=result_str)
 
-    def save_dims(self, *dims_boxes):
-        # first half of dims_boxes are dim names, second half are dim queries
-        all_dim_names = dims_boxes[: len(dims_boxes) // 2]
-        all_dim_queries = dims_boxes[len(dims_boxes) // 2 :]
-        self.dim_embeddings = {}
-        self.dim_queries = {}
-        for i in range(self.n_dims):
-            if all_dim_names[i].strip() != "":
-                dim_name = all_dim_names[i].strip()
-            else:  # if no name is given, use default name
-                dim_name = f"Dimension_{i+1}"
-            dim_queries = all_dim_queries[i].split("\n")  # dim queries
-            # remove empty queries or queries with only spaces
-            dim_queries = [q for q in dim_queries if q.strip() != ""]
-            dim_embedding = self.embed_texts(dim_queries)
-            mean_vect_dim = dim_embedding.mean(axis=0)
-            self.dim_embeddings[dim_name] = mean_vect_dim
-            self.dim_queries[dim_name] = dim_queries
-        # save self.dim_queries to json file
-        with open(Path(self.path_mgt.out_dir, "dimensions_queries.json"), "w") as f:
-            json.dump(self.dim_queries, f)
-        # update both positive and negative scale selectors
-        return (
-            [gr.Dropdown.update(choices=list(self.dim_queries.keys()))] * 20
-            + [
-                gr.Textbox.update(
-                    visible=True,
-                    value=f"Dimensions saved: {list(self.dim_queries.keys())}. You can download the json file below to keep a record of the final queries. Proceed to the next tab to define scales.",
-                )
-            ]
-            + [gr.File.update(visible=True)]
-            + [Path(self.path_mgt.out_dir, "dimensions_queries.json")]
-        )
+    def save_dims(self, progress=gr.Progress(), *dims_boxes):
+        try:
+            # first half of dims_boxes are dim names, second half are dim queries
+            all_dim_names = dims_boxes[: len(dims_boxes) // 2]
+            all_dim_queries = dims_boxes[len(dims_boxes) // 2 :]
+            self.dim_embeddings = {}
+            self.dim_queries = {}
+            for i in range(self.n_dims):
+                if all_dim_names[i].strip() != "":
+                    dim_name = all_dim_names[i].strip()
+                else:  # if no name is given, use default name
+                    dim_name = f"Dimension_{i+1}"
+                dim_queries = all_dim_queries[i].split("\n")  # dim queries
+                # remove empty queries or queries with only spaces
+                dim_queries = [q for q in dim_queries if q.strip() != ""]
+                dim_embedding = self.embed_texts(dim_queries, progress=progress)
+                mean_vect_dim = dim_embedding.mean(axis=0)
+                self.dim_embeddings[dim_name] = mean_vect_dim
+                self.dim_queries[dim_name] = dim_queries
+            # save self.dim_queries to json file
+            with open(Path(self.path_mgt.out_dir, "dimensions_queries.json"), "w") as f:
+                json.dump(self.dim_queries, f)
+            # update both positive and negative scale selectors
+            return (
+                [gr.Dropdown.update(choices=list(self.dim_queries.keys()))] * 20
+                + [
+                    gr.Textbox.update(
+                        visible=True,
+                        value=f"Dimensions saved: {list(self.dim_queries.keys())}. You can download the json file below to keep a record of the final queries. Proceed to the next tab to define scales.",
+                    )
+                ]
+                + [gr.File.update(visible=True)]
+                + [Path(self.path_mgt.out_dir, "dimensions_queries.json")]
+            )
+        except Exception as e:
+            return (
+                [gr.Dropdown.update(choices=list(self.dim_queries.keys()))] * 20
+                + [
+                    gr.Textbox.update(
+                        visible=True,
+                        value=f"Please check that you have filled in all the dimension names and queries. \n  Error message: {e}.",
+                    )
+                ]
+                + [gr.File.update(visible=False)]
+                + [None]
+            )
 
     def save_scales(self, *scale_boxes):
         # first 1/3 of scales boxes are names, second 1/3 are positive scales, third 1/3 are negative scales
-        all_scale_names = scale_boxes[: len(scale_boxes) // 3]
-        all_pos_scales = scale_boxes[len(scale_boxes) // 3 : 2 * len(scale_boxes) // 3]
-        all_neg_scales = scale_boxes[2 * len(scale_boxes) // 3 :]
-        self.scale_embeddings = {}
-        self.scale_definitions = {}
-        # all scale embeddings are average of positive scales subtracted by average of negative scales
-        for i in range(self.n_scales):
-            if all_scale_names[i].strip() != "":
-                scale_name = all_scale_names[i].strip()
-            else:  # if no name is given, use default name
-                scale_name = f"Scale_{i+1}"
-            # save scale definitions
-            """
-            example scale_definitions json format: 
-            scales = {
-                "External-Internal": {
-                    "Positive": ["Create", "Compete"],
-                    "Negative": ["Control", "Collaborate"],
-                },
-                "Flexible-Stable": {
-                    "Positive": ["Collaborate", "Create"],
-                    "Negative": ["Control", "Compete"],
-                },
-            }
-            """
-            self.scale_definitions[scale_name] = {
-                "Positive": all_pos_scales[i],
-                "Negative": all_neg_scales[i],
-            }
-            scale_embedding_pos = []
-            scale_embedding_neg = []
-            if len(all_pos_scales[i]) > 0:
-                for pos_scale in all_pos_scales[i]:
-                    scale_embedding_pos.append(self.dim_embeddings[pos_scale])
-            if len(all_neg_scales[i]) > 0:
-                for neg_scale in all_neg_scales[i]:
-                    scale_embedding_neg.append(self.dim_embeddings[neg_scale])
-            if (len(scale_embedding_pos) > 0) & (len(scale_embedding_neg) > 0):
-                # take averages and difference
-                scale_embedding = np.stack(scale_embedding_pos).mean(axis=0) - np.stack(
-                    scale_embedding_neg
-                ).mean(axis=0)
-            elif len(scale_embedding_pos) > 0:  # only positive scales
-                scale_embedding = np.stack(scale_embedding_pos).mean(axis=0)
-            elif len(scale_embedding_neg) > 0:  # only negative scales
-                scale_embedding = -np.stack(scale_embedding_neg).mean(axis=0)
-            else:  # no scales
-                pass
-            self.scale_embeddings[scale_name] = scale_embedding
-
-        # save self.scales to json file
-        with open(Path(self.path_mgt.out_dir, "scale_definitions.json"), "w") as f:
-            json.dump(self.scale_definitions, f)
-        return (
-            [
-                gr.Textbox.update(
-                    visible=True,
-                    value=f"Scales saved: {list(self.scale_definitions.keys())}. You can download the json file below to keep a record of the scale definitions. Proceed to the next tab to measure using semantic projection.",
-                )
+        try:
+            all_scale_names = scale_boxes[: len(scale_boxes) // 3]
+            all_pos_scales = scale_boxes[
+                len(scale_boxes) // 3 : 2 * len(scale_boxes) // 3
             ]
-            + [gr.File.update(visible=True)]
-            + [Path(self.path_mgt.out_dir, "scale_definitions.json")]
-        )
+            all_neg_scales = scale_boxes[2 * len(scale_boxes) // 3 :]
+            self.scale_embeddings = {}
+            self.scale_definitions = {}
+            # all scale embeddings are average of positive scales subtracted by average of negative scales
+            for i in range(self.n_scales):
+                if all_scale_names[i].strip() != "":
+                    scale_name = all_scale_names[i].strip()
+                else:  # if no name is given, use default name
+                    scale_name = f"Scale_{i+1}"
+                # save scale definitions
+                """
+                example scale_definitions json format: 
+                scales = {
+                    "External-Internal": {
+                        "Positive": ["Create", "Compete"],
+                        "Negative": ["Control", "Collaborate"],
+                    },
+                    "Flexible-Stable": {
+                        "Positive": ["Collaborate", "Create"],
+                        "Negative": ["Control", "Compete"],
+                    },
+                }
+                """
+                self.scale_definitions[scale_name] = {
+                    "Positive": all_pos_scales[i],
+                    "Negative": all_neg_scales[i],
+                }
+                scale_embedding_pos = []
+                scale_embedding_neg = []
+                if len(all_pos_scales[i]) > 0:
+                    for pos_scale in all_pos_scales[i]:
+                        scale_embedding_pos.append(self.dim_embeddings[pos_scale])
+                if len(all_neg_scales[i]) > 0:
+                    for neg_scale in all_neg_scales[i]:
+                        scale_embedding_neg.append(self.dim_embeddings[neg_scale])
+                if (len(scale_embedding_pos) > 0) & (len(scale_embedding_neg) > 0):
+                    # take averages and difference
+                    scale_embedding = np.stack(scale_embedding_pos).mean(
+                        axis=0
+                    ) - np.stack(scale_embedding_neg).mean(axis=0)
+                elif len(scale_embedding_pos) > 0:  # only positive scales
+                    scale_embedding = np.stack(scale_embedding_pos).mean(axis=0)
+                elif len(scale_embedding_neg) > 0:  # only negative scales
+                    scale_embedding = -np.stack(scale_embedding_neg).mean(axis=0)
+                else:  # no scales
+                    pass
+                self.scale_embeddings[scale_name] = scale_embedding
+
+            # save self.scales to json file
+            with open(Path(self.path_mgt.out_dir, "scale_definitions.json"), "w") as f:
+                json.dump(self.scale_definitions, f)
+
+            return (
+                [
+                    gr.Textbox.update(
+                        visible=True,
+                        value=f"Scales saved: {list(self.scale_definitions.keys())}. You can download the json file below to keep a record of the scale definitions. Proceed to the next tab to measure using semantic projection.",
+                    )
+                ]
+                + [gr.File.update(visible=True)]
+                + [Path(self.path_mgt.out_dir, "scale_definitions.json")]
+            )
+        except Exception as e:
+            # return error message in textbox update
+            return (
+                [
+                    gr.Textbox.update(
+                        visible=True,
+                        value=f"Error: Please check your scale definitions. Make sure that you have clicked the Embed Queries and Save Dimensions button in Tab 2, and there is no empty scales or dimensions. \nError message: {e}",
+                    )
+                ]
+                + [gr.File.update(visible=False)]
+                + [None]
+            )
 
     def measure_docs(self, whitening: str):
         scale_measures = []
         for scale in list(self.scale_definitions.keys()):
             measures = (
-                util.pytorch_cos_sim(self.embeddings, self.scale_embeddings[scale])
+                util_funcs.pytorch_cos_sim(
+                    self.embeddings, self.scale_embeddings[scale]
+                )
                 .cpu()
                 .numpy()
             )
@@ -467,6 +544,14 @@ class Meaurement:
                 gr.Row.update(visible=True),
                 Path(self.path_mgt.sample_data_dir, "sample_text.csv"),
                 Path(self.path_mgt.sample_data_dir, "sample_emb.npy"),
+                gr.Radio.update(value="Embed Documents"),
+                gr.Textbox.update(
+                    value="""Embedding completed! The shape of the embeddings is (2000, 384). You can download and save the embeddings below. Proceed to the next tab to define the dimensions."""
+                ),
+                gr.Dropdown.update(value="text"),
+                gr.Dropdown.update(value="doc_id"),
+                gr.Dropdown.update(value="Sentence Transformers (Local)"),
+                gr.Textbox.update(value="all-MiniLM-L6-v2"),
             ]
             + all_dim_name_boxes_updates
             + all_dim_seed_boxes_updates
@@ -477,9 +562,19 @@ class Meaurement:
         )
 
 
-def run_gui(out_dir="measure_output/", mode="local", username=None, password=None):
+def run_gui(
+    out_dir="measure_output/",
+    mode="local",
+    username=None,
+    password=None,
+    device="cpu",
+    **kwargs,
+):
     Path(out_dir).mkdir(parents=True, exist_ok=True)
     path_mgt = PathManager(out_dir=out_dir)
+
+    torch.set_grad_enabled(False)  # disable autograd
+
     with gr.Blocks(title="SPAR") as demo:
         m = Meaurement(path_mgt=path_mgt)
         gr.Markdown(
@@ -527,7 +622,7 @@ def run_gui(out_dir="measure_output/", mode="local", username=None, password=Non
         with gr.Tab("1. Upload File and Embed"):
             # read csv file and select columns
             gr.Markdown(
-                value=" 📖 Upload a CSV file that contains the text to be measured and a column that contains the document ID. Alternatively, you can click the Load Example Dataset and Scales button to explore the tool with an included sample dataset (4000 Facebook posts) and pre-defined Competing Values Framework (CVF) dimension and scales.",
+                value=" 📖 Upload a **CSV file** that contains at least 2 columns: (1) the documents to be measured, and (2) the document IDs. Alternatively, you can click the Load Example Dataset and Scales button on top to explore the tool with an included sample dataset (4000 Facebook posts) and pre-defined Competing Values Framework (CVF) dimension and scales.",
                 label="",
             )
             with gr.Row(variant="panel"):
@@ -567,17 +662,18 @@ def run_gui(out_dir="measure_output/", mode="local", username=None, password=Non
                         "Sentence Transformers (Local)",
                         "OpenAI text-embedding-ada-002 (API Key Required)",
                     ],
-                    label="Select an embedding model:",
+                    label="Select an embedding method:",
                     interactive=True,
                     value="Sentence Transformers (Local)",
                     multiselect=False,
                 )
                 openai_api_key = gr.Textbox(
-                    placeholder="We strongly recommed revoking your API key after public testing as our server does not provide any security guarantees to prevent potential security breaches.",
+                    placeholder="",
                     label="OpenAI API Key (leave blank if key is set as Environment Variable)",
                     value="",
                     visible=False,
                     interactive=True,
+                    type="password",
                 )
                 sbert_model_textbox = gr.Textbox(
                     label="Sentence Transformers Model Name",
@@ -615,7 +711,7 @@ def run_gui(out_dir="measure_output/", mode="local", username=None, password=Non
                 )
                 emb_result_txtbox = gr.Textbox(
                     value="",
-                    label="Embedding Progress (check console for progress updates)",
+                    label="Embedding Progress",
                 )
                 emb_results_file = gr.File(visible=False)
 
@@ -659,7 +755,7 @@ def run_gui(out_dir="measure_output/", mode="local", username=None, password=Non
             )
         with gr.Tab("2. Define Dimensions and Semantic Search"):
             gr.Markdown(
-                value=" 📖 Move the sliders to set the number of dimensions and the number of results in each round of semantic search. ",
+                value=" 📖 Move the sliders to set the number of dimensions and the number of results in each round of semantic search. Make sure that there is no empty dimension. ",
                 label="",
             )
             n_dim_slider = gr.Slider(
@@ -908,6 +1004,12 @@ def run_gui(out_dir="measure_output/", mode="local", username=None, password=Non
                 example_row,
                 example_file_download,
                 example_emb_download,
+                upload_emb_choice,
+                emb_result_txtbox,
+                doc_col_selector,
+                doc_id_col_selector,
+                embedding_model_dropdown,
+                sbert_model_textbox,
             ]
             + all_dim_name_boxes
             + all_search_query_boxes
@@ -917,21 +1019,27 @@ def run_gui(out_dir="measure_output/", mode="local", username=None, password=Non
             + [n_dim_slider, n_scale_slider],
         )
     demo.queue(concurrency_count=3)
+    if username is None or password is None:
+        auth = None
+    else:
+        auth = (username, password)
     if mode == "public":
         demo.launch(
             share=True,
-            auth=(username, password),
+            auth=auth,
             server_name="0.0.0.0",
-            favicon_path="sample_data/favicon.png",
+            favicon_path=path_mgt.sample_data_dir / "favicon.png",
+            **kwargs,
         )
     elif mode == "intranet":
         demo.launch(
             server_name="0.0.0.0",
-            auth=(username, password),
-            favicon_path="sample_data/favicon.png",
+            auth=auth,
+            favicon_path=path_mgt.sample_data_dir / "favicon.png",
+            **kwargs,
         )
     elif mode == "local":
-        demo.launch(favicon_path="sample_data/favicon.png")
+        demo.launch(favicon_path=path_mgt.sample_data_dir / "favicon.png", **kwargs)
     else:
         print("Invalid mode. Please choose from 'public', 'intranet', or 'local'.")
 
